
const char *sffs_fs_c_version = "$Id: sffs_fs.c,SFFS 2007/11/16 $";
extern const char *sffs_guts_c_version;


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/pagemap.h>
#include <linux/mtd/mtd.h>
#include <linux/interrupt.h>
#include <linux/string.h>
#include <linux/locks.h>

#include <asm/uaccess.h>

#include "yportenv.h"
#include "sffs_guts.h"


#ifdef KERN_DEBUG
#define KERN_DEBUG "<6>"
#endif


unsigned sffs_traceMask = SFFS_TRACE_ALWAYS | SFFS_TRACE_BAD_BLOCKS;

#ifdef CONFIG_SFFS_RAM_ENABLED
#include "sffs_nandemul.h" 
// 2 MB of RAM for emulation
#define SFFS_RAM_EMULATION_SIZE  0x200000
#endif //CONFIG_SFFS_RAM_ENABLED

#ifdef CONFIG_SFFS_MTD_ENABLED
#include <linux/mtd/mtd.h>
#include "sffs_mtdif.h"
#endif //CONFIG_SFFS_MTD_ENABLED

//#define T(x) printk x



#define sffs_InodeToObject(iptr) ((sffs_Object *)((iptr)->u.generic_ip))
#define sffs_DentryToObject(dptr) sffs_InodeToObject((dptr)->d_inode)
#define sffs_SuperToDevice(sb)	((sffs_Device *)sb->u.generic_sbp)



static void sffs_put_super(struct super_block *sb);

static ssize_t sffs_file_read(struct file *f, char *buf, size_t n, loff_t *pos);
static ssize_t sffs_file_write(struct file *f, const char *buf, size_t n, loff_t *pos);
static int sffs_file_flush(struct file* file);

static int sffs_sync_object(struct file * file, struct dentry *dentry, int datasync);

static int sffs_readdir(struct file *f, void *dirent, filldir_t filldir);

static int sffs_create(struct inode *dir, struct dentry *dentry, int mode);
static struct dentry * sffs_lookup(struct inode *dir, struct dentry *dentry);
static int sffs_link(struct dentry *old_dentry, struct inode * dir, struct dentry * dentry);
static int sffs_unlink(struct inode * dir, struct dentry *dentry);
static int sffs_symlink(struct inode * dir, struct dentry *dentry, const char * symname);
static int sffs_mkdir(struct inode * dir, struct dentry * dentry, int mode);
static int sffs_mknod(struct inode *dir, struct dentry *dentry, int mode, int dev);
static int sffs_rename(struct inode * old_dir, struct dentry *old_dentry, struct inode * new_dir,struct dentry *new_dentry);
static int sffs_setattr(struct dentry *dentry, struct iattr *attr);

static int sffs_statfs(struct super_block *sb, struct statfs *buf);
static void sffs_read_inode (struct inode *inode);
static struct super_block *sffs_read_super(struct super_block * sb, void * data, int silent);
static void sffs_put_inode (struct inode *inode);
static void sffs_delete_inode(struct inode *);
static void sffs_clear_inode(struct inode *);

static int sffs_readpage(struct file *file, struct page * page);

static int sffs_prepare_write(struct file *f, struct page *pg, unsigned offset, unsigned to);
static int sffs_commit_write(struct file *f, struct page *pg, unsigned offset, unsigned to);

static int sffs_readlink(struct dentry *dentry, char *buffer, int buflen);
static int sffs_follow_link(struct dentry *dentry, struct nameidata *nd);




static struct address_space_operations sffs_file_address_operations = {
	readpage:		sffs_readpage,
	prepare_write:	sffs_prepare_write,
	commit_write:	sffs_commit_write
};

static struct file_operations sffs_file_operations = {

#ifdef CONFIG_SFFS_USE_GENERIC_RW
	read:		generic_file_read,
	write:		generic_file_write,
#else
	read:		sffs_file_read,//
	write:		sffs_file_write,//
#endif
	mmap:		generic_file_mmap,
	flush:		sffs_file_flush,//
	fsync:		sffs_sync_object,
};

static struct inode_operations sffs_file_inode_operations = {
	setattr:	sffs_setattr,
};

struct inode_operations sffs_symlink_inode_operations =
{	
	follow_link:	sffs_follow_link,
	setattr:	sffs_setattr
};


static struct inode_operations sffs_dir_inode_operations = {
	create:		sffs_create,//
	lookup:		sffs_lookup,//

	link:		sffs_link,//
	unlink:		sffs_unlink,	//
	symlink:	sffs_symlink,
	
	mkdir:		sffs_mkdir,//
	rmdir:		sffs_unlink,
	mknod:		sffs_mknod,//
	rename:		sffs_rename,
	setattr:	sffs_setattr,
};

static struct file_operations sffs_dir_operations = {
	read:		generic_read_dir,
	readdir:	sffs_readdir,//
	fsync:		sffs_sync_object,
};


static struct super_operations sffs_super_ops = {
	statfs:			sffs_statfs,//x
	read_inode:		sffs_read_inode,
	put_inode:		sffs_put_inode,
	put_super:		sffs_put_super,
	delete_inode:		sffs_delete_inode,
	clear_inode:		sffs_clear_inode,
};



static void sffs_GrossLock(sffs_Device *dev)
{
	down(&dev->grossLock);
}

static void sffs_GrossUnlock(sffs_Device *dev)
{
	up(&dev->grossLock);
}

static int sffs_readlink(struct dentry *dentry, char *buffer, int buflen)
{
	unsigned char *alias;
	int ret;

	sffs_Device *dev = sffs_DentryToObject(dentry)->myDev;


	sffs_GrossLock(dev);
	
	alias = sffs_GetSymlinkAlias(sffs_DentryToObject(dentry));
	
	sffs_GrossUnlock(dev);
	
	if(!alias)
		return -ENOMEM;

	ret = vfs_readlink(dentry, buffer, buflen, alias);
	kfree(alias);
	return ret;
}

static int sffs_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	unsigned char *alias;
	int ret;
	sffs_Device *dev = sffs_DentryToObject(dentry)->myDev; 

	sffs_GrossLock(dev);

	alias = sffs_GetSymlinkAlias(sffs_DentryToObject(dentry));
	
	sffs_GrossUnlock(dev);
	
	if(!alias)
		return -ENOMEM;

	ret = vfs_follow_link(nd,alias);
	kfree(alias);
	return ret;
}


struct inode *sffs_get_inode(struct super_block *sb, int mode, int dev,sffs_Object *obj);

/*
 * Lookup is used to find objects in the fs
 */
 static struct dentry * sffs_lookup(struct inode *dir, struct dentry *dentry)
{
	sffs_Object *obj;
	struct inode *inode;
	
	sffs_Device *dev = sffs_InodeToObject(dir)->myDev;


	sffs_GrossLock(dev);

		
	obj = sffs_FindObjectByName(sffs_InodeToObject(dir),dentry->d_name.name);
	
	obj = sffs_GetEquivalentObject(obj); // in case it was a hardlink
	

	
	if(obj)
	{
		
		inode = sffs_get_inode(dir->i_sb, obj->st_mode,0,obj);
		
		if(inode)
		{
 			d_add(dentry,inode);
			
			sffs_GrossUnlock(dev);
 
			return NULL;
		}

	}
	else
	{

		
	}
	sffs_GrossUnlock(dev);
	
	return NULL;

	
}

// For now put inode is just for debugging
// Put inode is called when the inode **structure** is put.
static void sffs_put_inode(struct inode *inode)
{
}

// clear is called to tell the fs to release any per-inode data it holds
static void sffs_clear_inode(struct inode *inode)
{
	sffs_Object *obj = sffs_InodeToObject(inode);

 
	if(obj)
	{
		obj->myInode = NULL;
		inode->u.generic_ip = NULL;	
	}
	
	
}

// delete is called when the link count is zero and the inode
// is put (ie. nobody wants to know about it anymore, time to
// delete the file).
// NB Must call clear_inode()
static void sffs_delete_inode(struct inode *inode)
{
	sffs_Object *obj = sffs_InodeToObject(inode);
	sffs_Device *dev;
	
	if(obj)
	{
	 	dev = obj->myDev;
		sffs_GrossLock(dev);
		sffs_DeleteFile(obj);
		sffs_GrossUnlock(dev);
	}
	clear_inode(inode);
}


 static int sffs_file_flush(struct file* file)
{
    int i;
//    __u8 data[512];
//    sffs_Spare spare;
    
	sffs_Object *obj = sffs_DentryToObject(file->f_dentry);
	
	sffs_Device *dev = obj->myDev;

 
	sffs_GrossLock(dev);
    


    g_nCacheCnt=0;
    sffs_FlushFile(obj,1); 
	sffs_GrossUnlock(dev);

 
    g_nCreateflag=0;
     i=0;
 
    g_nCacheCnt=0;    

    return 0;
}



 static int sffs_readpage_nolock(struct file *f, struct page * pg)
{
	
	sffs_Object *obj;
	unsigned char *pg_buf;	
	int ret;
    int size;
	
	sffs_Device *dev;
 	obj  = sffs_DentryToObject(f->f_dentry);

	dev = obj->myDev;
	
	
	if (!PageLocked(pg))
                PAGE_BUG(pg);

	pg_buf = kmap(pg);


	sffs_GrossLock(dev);
	ret = sffs_ReadDataFromFile(obj, pg_buf, pg->index << PAGE_CACHE_SHIFT, PAGE_CACHE_SIZE);

    size=obj->variant.fileVariant.fileSize;
    size=size%PAGE_CACHE_SIZE;
    size=(obj->variant.fileVariant.fileSize)-(pg->index << PAGE_CACHE_SHIFT);
    size=(obj->variant.fileVariant.fileSize)%PAGE_CACHE_SIZE;
    
     if((obj->variant.fileVariant.fileSize)-(pg->index << PAGE_CACHE_SHIFT)<=size)
    {
            g_nStartRead=0;    
//            do_gettimeofday(&g_ReadEndTime);        
  //          g_ltemp= (g_ReadEndTime.tv_sec * 1000000 + g_ReadEndTime.tv_usec) - (g_ReadStartTime.tv_sec * 1000000 + g_ReadStartTime.tv_usec);

    }

	sffs_GrossUnlock(dev);
	
	if(ret >= 0) ret = 0;

	if (ret) {
		ClearPageUptodate(pg);
		SetPageError(pg);
	} else {
		SetPageUptodate(pg);
		ClearPageError(pg);
	}

	flush_dcache_page(pg);
	kunmap(pg);
 
	return ret;
}

static int sffs_readpage_unlock(struct file *f, struct page *pg)
{
	int ret = sffs_readpage_nolock(f,pg);
	UnlockPage(pg);
	return ret;
}

static int sffs_readpage(struct file *f, struct page * pg)
{
    int rvalue;
    
     if(g_nStartRead==0)
    {
//        do_gettimeofday(&g_ReadStartTime);    
        g_nStartRead=1;
    }
 
    
    rvalue=sffs_readpage_unlock(f,pg);
 	return rvalue;
}
 
static int sffs_prepare_write(struct file *f, struct page *pg, unsigned offset, unsigned to)
{
 
	sffs_Object *obj;
//	int ret;
//    int m_nChunk;
	sffs_Device *dev;
	obj  = sffs_DentryToObject(f->f_dentry);

	dev = obj->myDev;

    if(g_nCrashStart!=1)
    {
// 	    do_gettimeofday(&g_WriteSartTime);    
        
        if(g_nCreateflag!=1)
            sffs_TransactionStart(dev,1);
        else
            sffs_TranactionCreateStart(dev,g_nFileCreate); 
        g_nCrashStart=1;
    }
   
	if(!Page_Uptodate(pg) && (offset || to < PAGE_CACHE_SIZE))
		return  sffs_readpage_nolock(f,pg);    

	return 0;
	
}

static int sffs_commit_write(struct file *f, struct page *pg, unsigned offset, unsigned to)
{

	void *addr = page_address(pg) + offset;
	loff_t pos = (((loff_t)pg->index) << PAGE_CACHE_SHIFT) + offset;
	int nBytes = to - offset;
	int nWritten;
	
	unsigned spos = pos;
	unsigned saddr = (unsigned)addr;

   
	
	nWritten = sffs_file_write(f,addr, nBytes, &pos);

         
    if(nWritten != nBytes)
	{ 
		SetPageError(pg);
		ClearPageUptodate(pg);
	}
	else
	{
		SetPageUptodate(pg);
	}
 
	return nWritten;

}



//sffs_read_inode() -> sffs_FillInodeFromObject()
//Ʈ vfs(?) read_inode ϰ ű⼭  Լ ȣ.
//inode dentry, file, addrsee_space ü( Լ   ͵ ˷ش.)
//ls ׷ ɾ ÿ ϸ ȣ.
//ext2 Ͻý ȿ inodeͷ Ϳ Ǿ.

//sffs inode ƴ object  ǰ ⼭ Ư 丮  ҽ
//inode ͸ Ѵ.
/*
sffs_FillInodeFromObject
	sffs_GetObjectFileLength
	sffs_GetObjectLinkCount
*/
static void sffs_FillInodeFromObject(struct inode *inode, sffs_Object *obj)
{
	if (inode && obj) 
	{
		inode->i_ino = obj->objectId;
		inode->i_mode = obj->st_mode;
		inode->i_uid = obj->st_uid;
		inode->i_gid = obj->st_gid;
		inode->i_blksize = inode->i_sb->s_blocksize;
		inode->i_blocks = 0;
		inode->i_rdev = obj->st_rdev;;
		inode->i_atime = obj->st_atime;
		inode->i_mtime = obj->st_mtime;
		inode->i_ctime = obj->st_ctime;
		inode->i_size = sffs_GetObjectFileLength(obj);
		inode->i_nlink = sffs_GetObjectLinkCount(obj);
		
 		switch (obj->st_mode & S_IFMT) 
		{
			default: // fifo, device or socket
				init_special_inode(inode, obj->st_mode, obj->st_rdev);
				break;
			case S_IFREG:	// file		
				inode->i_op = &sffs_file_inode_operations;
				inode->i_fop = &sffs_file_operations;
				inode->i_mapping->a_ops = &sffs_file_address_operations;
				break;
			case S_IFDIR:	// directory 
				inode->i_op = &sffs_dir_inode_operations;
				inode->i_fop = &sffs_dir_operations;
				break;
			case S_IFLNK:	// symlink 
				inode->i_op = &sffs_symlink_inode_operations;
				break;
		} 
		
		inode->u.generic_ip = obj;
		obj->myInode = inode;
		
	}
	else
	{
 
	}

}

struct inode *sffs_get_inode(struct super_block *sb, int mode, int dev,sffs_Object *obj)
{
	struct inode * inode;
	
 

	inode = iget(sb,obj->objectId);
 	
	return inode;
}

 
static ssize_t sffs_file_read(struct file *f, char *buf, size_t n, loff_t *pos)
{
	sffs_Object *obj;
	int nRead,ipos;
	struct inode *inode;
	sffs_Device *dev;


 
	obj  = sffs_DentryToObject(f->f_dentry);
	
	dev = obj->myDev;
	
	sffs_GrossLock(dev);
	
	inode = f->f_dentry->d_inode;
	
	if (*pos < inode->i_size) 
	{
			if (*pos + n > inode->i_size)
			{
				n = inode->i_size - *pos;
			}
	}
	else
	{
		n = 0;
	}
	
	nRead = sffs_ReadDataFromFile(obj,buf,*pos,n);
	if(nRead > 0)
	{
		f->f_pos += nRead;
	}
	
	sffs_GrossUnlock(dev);
	
	ipos = *pos;
	
 
    
	return nRead;
	
}


static ssize_t sffs_file_write(struct file *f, const char *buf, size_t n, loff_t *pos)
{
	sffs_Object *obj;
	int nWritten,ipos;
	struct inode *inode;
	sffs_Device *dev;
	
 
	obj  = sffs_DentryToObject(f->f_dentry);
	
	dev = obj->myDev;
	
	sffs_GrossLock(dev);
	
	inode = f->f_dentry->d_inode;

	if(!S_ISBLK(inode->i_mode) && f->f_flags & O_APPEND)
	{
		ipos = inode->i_size;
	}
	else
	{
		ipos = *pos;
	}
	
	
	if(!obj)
	{
 
	}
	else
	{
 
	}

 	nWritten = sffs_WriteDataToFile(obj,buf,ipos,n);

 
	if(nWritten > 0)
	{
		ipos += nWritten;
		*pos = ipos;
		if(ipos > inode->i_size)
		{
			inode->i_size = ipos;
			inode->i_blocks = (ipos + inode->i_blksize - 1)/ inode->i_blksize;
			 
		}
		
	}
	sffs_GrossUnlock(dev);
	
	return nWritten != n ? -ENOSPC : nWritten;	
}


static int sffs_readdir(struct file *f, void *dirent, filldir_t filldir)
{
	sffs_Object *obj;
	sffs_Device *dev;
	struct inode *inode = f->f_dentry->d_inode;
	unsigned long offset, curoffs;
	struct list_head *i;	
	sffs_Object *l;
	
	char name[SFFS_MAX_NAME_LENGTH	+1];
		
	obj =  sffs_DentryToObject(f->f_dentry);
	dev = obj->myDev;
	
	sffs_GrossLock(dev);
	
	offset = f->f_pos;
	 
	
	if(offset == 0)
	{ 
		if(filldir(dirent,".",1,offset,inode->i_ino,DT_DIR) < 0)
		{
			goto out;
		}
		offset++;
		f->f_pos++;
	}
	if(offset == 1)
	{ 
		if(filldir(dirent,"..",2,offset,f->f_dentry->d_parent->d_inode->i_ino,DT_DIR) < 0)
		{
			goto out;
		}
		offset++;
		f->f_pos++;
	}
	
	curoffs = 1;
	
	list_for_each(i,&obj->variant.directoryVariant.children)
	{
		curoffs++;
		if(curoffs >= offset)
		{		
			l = list_entry(i, sffs_Object,siblings);
			
			sffs_GetObjectName(l,name,SFFS_MAX_NAME_LENGTH+1);  
			
			if(filldir(dirent,
					   name,
					   strlen(name),
					   offset,
					   sffs_GetObjectInode(l),
					   sffs_GetObjectType(l))
					   < 0)
			{
				goto up_and_out;
			}
			
			offset++;
			f->f_pos++;	   
		}
	}

  up_and_out:
  out:
  
    sffs_GrossUnlock(dev);
    
	return 0;
}


// ο Ư   inode , dentry  mode( ), rdev(ġ major ȣ) 
static int sffs_mknod(struct inode *dir, struct dentry *dentry, int mode, int rdev)
{
	struct inode *inode;
	
	sffs_Object *obj = NULL;
	sffs_Device *dev;
 
	sffs_Object *parent = sffs_InodeToObject(dir);
	
	int error = -ENOSPC;
        
	if(parent)
	{
	}
	else
	{
		return -EPERM;
	}
	

	dev = parent->myDev;
	
	sffs_GrossLock(dev);

	switch (mode & S_IFMT) 
	{
		//Ư Ѵ. char device, block device
		default:
			obj = sffs_MknodSpecial(parent,dentry->d_name.name,mode,current->uid, current->gid,rdev);
			break;
		case S_IFREG:	// file		
            g_nCreateflag=1;
			obj = sffs_MknodFile(parent,dentry->d_name.name,mode,current->uid, current->gid);
			break;
		case S_IFDIR:	// directory
			obj = sffs_MknodDirectory(parent,dentry->d_name.name,mode,current->uid, current->gid);
			break;
		case S_IFLNK:	// symlink
			obj = NULL; 
			break;
	}
	
	if(obj)
	{
		inode = sffs_get_inode(dir->i_sb, mode, rdev, obj);
		d_instantiate(dentry, inode);
		error = 0;
	}
	else
	{
		error = -ENOMEM;
	}

	sffs_GrossUnlock(dev);

	return error;
}

static int sffs_mkdir(struct inode * dir, struct dentry * dentry, int mode)
{
	int retVal;
	retVal =  sffs_mknod(dir, dentry, mode | S_IFDIR, 0);
#if 0
	if(!retVal)
	{
		dget(dentry);
	}
#endif
	return retVal;
}

static int sffs_create(struct inode *dir, struct dentry *dentry, int mode)
{
	return sffs_mknod(dir, dentry, mode | S_IFREG, 0);
}


static int sffs_unlink(struct inode * dir, struct dentry *dentry)
{
	int retVal;
	
	sffs_Device *dev;
	
	dev = sffs_InodeToObject(dir)->myDev;
	
	sffs_GrossLock(dev);
	
	
	retVal = sffs_Unlink(sffs_InodeToObject(dir),dentry->d_name.name);
	
	
	sffs_GrossUnlock(dev);
	
	if( retVal == SFFS_OK)
	{
		dentry->d_inode->i_nlink--;
		mark_inode_dirty(dentry->d_inode);
		return 0;
	}
	else
	{
		return -ENOTEMPTY;
	}
}


static int sffs_link(struct dentry *old_dentry, struct inode * dir, struct dentry * dentry)
{
	struct inode *inode = old_dentry->d_inode;
	sffs_Object *obj = NULL;
	sffs_Object *link=NULL;
	sffs_Device *dev;
	
	obj = sffs_InodeToObject(inode);
	dev = obj->myDev;
	
	sffs_GrossLock(dev);

	if (!S_ISDIR(inode->i_mode))	// Don't link directories
	{
		link = sffs_Link(sffs_InodeToObject(dir),dentry->d_name.name,obj);
	}
	

	if(link)
	{
		old_dentry->d_inode->i_nlink =  sffs_GetObjectLinkCount(obj);
		d_instantiate(dentry, old_dentry->d_inode);
		atomic_inc(&old_dentry->d_inode->i_count);
	
	}
	
	sffs_GrossUnlock(dev);
	

	if(link)
	{
	
		return 0;
	}
	
	
	return -EPERM; 
}


static int sffs_symlink(struct inode * dir, struct dentry *dentry, const char * symname)
{
	sffs_Object *obj;
	sffs_Device *dev;
	

	dev = sffs_InodeToObject(dir)->myDev;
	sffs_GrossLock(dev);
	obj = sffs_MknodSymLink(sffs_InodeToObject(dir), dentry->d_name.name, 
							 S_IFLNK | S_IRWXUGO, current->uid, current->gid,
							 symname);
	sffs_GrossUnlock(dev);

	if(obj)
	{

		struct inode* inode;
	
		inode = sffs_get_inode(dir->i_sb, obj->st_mode, 0, obj);
		d_instantiate(dentry, inode);
		return 0;
	}
	else
	{

	}
	
	return -ENOMEM;
}

static int sffs_sync_object(struct file * file, struct dentry *dentry, int datasync)
{

	sffs_Object *obj;
	sffs_Device *dev;

	obj = sffs_DentryToObject(dentry);

	dev = obj->myDev;
	
	sffs_GrossLock(dev);
	sffs_FlushFile(obj,1);
	sffs_GrossUnlock(dev);
	return 0;
}

static int sffs_rename(struct inode * old_dir, struct dentry *old_dentry, struct inode * new_dir,struct dentry *new_dentry)
{
	sffs_Device *dev;
	int retVal = SFFS_FAIL;
	int removed = 0;
	sffs_Object *target;
	
	dev = sffs_InodeToObject(old_dir)->myDev;

	sffs_GrossLock(dev);
	
	target = sffs_FindObjectByName(sffs_InodeToObject(new_dir),new_dentry->d_name.name);
	
	if(target &&
	   target->variantType == SFFS_OBJECT_TYPE_DIRECTORY &&
	   !list_empty(&target->variant.directoryVariant.children))
	{
		retVal = SFFS_FAIL;
	}
	else
	{
	   
		// Unlink the target if it exists
		removed = sffs_Unlink(sffs_InodeToObject(new_dir),new_dentry->d_name.name);

	
		retVal =  sffs_RenameObject(sffs_InodeToObject(old_dir),old_dentry->d_name.name,
									sffs_InodeToObject(new_dir),new_dentry->d_name.name);
									
	}
	sffs_GrossUnlock(dev);
	
	if(retVal == SFFS_OK)
	{
		if(removed == SFFS_OK)
		{
			new_dentry->d_inode->i_nlink--;
			mark_inode_dirty(new_dentry->d_inode);
		}
		
		return 0;
	}
	else
	{
		return -ENOTEMPTY;
	}
	

}

static int sffs_setattr(struct dentry *dentry, struct iattr *attr)
{
	struct inode *inode = dentry->d_inode;
	int error;
	sffs_Device *dev;
	
	
	if((error = inode_change_ok(inode,attr)) == 0)
	{
	
		dev = sffs_InodeToObject(inode)->myDev;
		sffs_GrossLock(dev);
		if(sffs_SetAttributes(sffs_InodeToObject(inode),attr) == SFFS_OK)
		{
			error = 0;
		}
		else
		{
			error = -EPERM;
		}
		sffs_GrossUnlock(dev);
		inode_setattr(inode,attr);
	}
    g_nTranFlag=0;
	return error;
}

static int sffs_statfs(struct super_block *sb, struct statfs *buf)
{
	sffs_Device *dev = sffs_SuperToDevice(sb);
    
	sffs_GrossLock(dev);
	
	buf->f_type = SFFS_MAGIC;
	buf->f_bsize = sb->s_blocksize;
	buf->f_namelen = 255;
	buf->f_blocks = (dev->endBlock - dev->startBlock + 1) * SFFS_CHUNKS_PER_BLOCK/
						(sb->s_blocksize/SFFS_BYTES_PER_CHUNK);
	buf->f_files = 0;
	buf->f_ffree = 0;
	buf->f_bfree = sffs_GetNumberOfFreeChunks(dev)/
						(sb->s_blocksize/SFFS_BYTES_PER_CHUNK);
	buf->f_bavail =  buf->f_bfree;
	
	sffs_GrossUnlock(dev);
	return 0;
}

static void sffs_read_inode (struct inode *inode)
{
	// NB This is called as a side effect of other functions and
	// thus gross locking should always be in place already.
	
	sffs_Object *obj ; 
	sffs_Device *dev = sffs_SuperToDevice(inode->i_sb);

    

	obj  = sffs_FindObjectByNumber(dev,inode->i_ino);

    debug_obect(obj);

	sffs_FillInodeFromObject(inode,obj);

}



static sffs_Device *sffs_dev;
static sffs_Device *sffsram_dev;



static void sffs_put_super(struct super_block *sb)
{
	sffs_Device *dev;
    dev= sffs_SuperToDevice(sb);


    
	sffs_GrossLock(dev);
	if(dev->putSuperFunc)
	{
		 dev->putSuperFunc(sb);
	}
	sffs_Deinitialise(dev);
	sffs_GrossUnlock(dev);

	if(dev == sffs_dev) sffs_dev = NULL;
	if(dev == sffsram_dev) sffsram_dev = NULL;
		
	kfree(dev);
}


#ifdef CONFIG_SFFS_MTD_ENABLED

static void  sffs_MTDPutSuper(struct super_block *sb)
{
	
	struct mtd_info *mtd = sffs_SuperToDevice(sb)->genericDevice;

	if(mtd->sync)
	{
		mtd->sync(mtd);
	}
	
	put_mtd_device(mtd);
}

#endif


//sffs_read_super ȣ. 
//mount vfs sb .
//sffs_device ü ְ sffs mtd Ѵٴ° .
//get_mtd mtd ̽ Ѵٰ .
static struct super_block *sffs_internal_read_super(int useRam, struct super_block * sb, void * data, int silent)
{
	int nBlocks;
	struct inode * inode;
	struct dentry * root;
	sffs_Device *dev;
	
	sb->s_magic = SFFS_MAGIC;
	sb->s_op = &sffs_super_ops;
    printk("mount : sffs_internal_read_super\n");
	

//init_sffs Ͻý ϸ ϰ ⼭ ƮҴ sb  .
#ifdef CONFIG_SFFS_USE_CHUNK_SIZE
	sb->s_blocksize = SFFS_BYTES_PER_CHUNK;
	sb->s_blocksize_bits = SFFS_CHUNK_SIZE_SHIFT;
#else
	sb->s_blocksize = PAGE_CACHE_SIZE;
	sb->s_blocksize_bits = PAGE_CACHE_SHIFT;
#endif

#ifdef CONFIG_SFFS_DISABLE_WRITE_VERIFY

#endif


	if(useRam)
	{

#ifdef CONFIG_SFFS_RAM_ENABLED
		// Set the sffs_Device up for ram emulation

		sb->u.generic_sbp =	dev = kmalloc(sizeof(sffs_Device),GFP_KERNEL);
		if(!dev)
		{
			// Deep shit could not allocate device structure
			//printk(KERN_DEBUG"sffs_read_super: Failed trying to allocate sffs_Device. Terminating debug.\n");
			return NULL;
		}

		memset(dev,0,sizeof(sffs_Device));
		dev->genericDevice = NULL; // Not used for RAM emulation.

		nBlocks = SFFS_RAM_EMULATION_SIZE / (SFFS_CHUNKS_PER_BLOCK * SFFS_BYTES_PER_CHUNK);
		dev->startBlock = 1;  // Don't use block 0
		dev->endBlock = nBlocks - 1;
		dev->nChunksPerBlock = SFFS_CHUNKS_PER_BLOCK;
		dev->nBytesPerChunk = SFFS_BYTES_PER_CHUNK;
		dev->nReservedBlocks = 5;

		dev->writeChunkToNAND = nandemul_WriteChunkToNAND;
		dev->readChunkFromNAND = nandemul_ReadChunkFromNAND;
		dev->eraseBlockInNAND = nandemul_EraseBlockInNAND;
		dev->initialiseNAND = nandemul_InitialiseNAND;

		sffsram_dev = dev;
		
#endif

	}
    //MTD
	else
	{	
#ifdef CONFIG_SFFS_MTD_ENABLED
		struct mtd_info *mtd;
		
		//mtdڵ ´.
		mtd = get_mtd_device(NULL, MINOR(sb->s_dev));
		if (!mtd) 
		{
			return NULL;
		}
		
		if(mtd->type != MTD_NANDFLASH)
		{
			return NULL;
		}



		if(!mtd->erase ||
		   !mtd->read  ||
		   !mtd->write ||
#ifndef CONFIG_SFFS_USE_OLD_MTD
		   !mtd->write_ecc ||
		   !mtd->read_ecc ||
#endif
		   !mtd->read_oob ||
		   !mtd->write_oob )
		{
			return NULL;
		}
		
		if(mtd->oobblock != SFFS_BYTES_PER_CHUNK ||
		   mtd->oobsize != SFFS_BYTES_PER_SPARE)
		{
			return NULL;
		}
		   

		// OK, so if we got here, we have an MTD that's NAND and looks 
		// like it has the right capabilities
		// Set the sffs_Device up for ram emulation

        //sffs superblock ⺻  ʴ° . ׷ Ѵ.
		sb->u.generic_sbp =	dev = kmalloc(sizeof(sffs_Device),GFP_KERNEL);
		if(!dev)
		{
			// Deep shit could not allocate device structure
			return NULL;
		}

        //sffs sffs_device ü äش.
		memset(dev,0,sizeof(sffs_Device));
		dev->genericDevice = mtd; 

		// Set up the memory size parameters....


		nBlocks = mtd->size / (SFFS_CHUNKS_PER_BLOCK * SFFS_BYTES_PER_CHUNK);
		dev->startBlock = 1;  // Don't use block 0
		dev->endBlock = nBlocks - 1;
		dev->nChunksPerBlock = SFFS_CHUNKS_PER_BLOCK;
		dev->nBytesPerChunk = SFFS_BYTES_PER_CHUNK;
		dev->nReservedBlocks = 5;
		dev->nShortOpCaches = 10; // Enable short op caching
		
        
        //mtd .
		dev->writeChunkToNAND = nandmtd_WriteChunkToNAND;
		dev->readChunkFromNAND = nandmtd_ReadChunkFromNAND;
		dev->eraseBlockInNAND = nandmtd_EraseBlockInNAND;
		dev->initialiseNAND = nandmtd_InitialiseNAND;

                
		dev->putSuperFunc = sffs_MTDPutSuper;
		
#ifdef CONFIG_SFFS_USE_NANDECC
		dev->useNANDECC = 1;
#endif

		sffs_dev = dev;
		
#endif
	}

	init_MUTEX(&dev->grossLock);
	
	
	sffs_GrossLock(dev);
    //Ʈ  丮  ֱ  Ʈ Ѵ.
    //inode dentry о ʱ  Ѵ.
    //Tnode  ÷ ?
	sffs_GutsInitialise(sffs_SuperToDevice(sb));

	// Create root inode
	inode = sffs_get_inode(sb, S_IFDIR | 0755, 0,sffs_Root(sffs_SuperToDevice(sb)));

	sffs_GrossUnlock(dev);

	if (!inode)
		return NULL;


	root = d_alloc_root(inode);

	if (!root) {
		iput(inode);
		return NULL;
	}
	sb->s_root = root;

	return sb;
}

#ifdef CONFIG_SFFS_MTD_ENABLED

//Ͻý Ʈ Ҷ ȣ.
static struct super_block *sffs_read_super(struct super_block * sb, void * data, int silent)
{
	return sffs_internal_read_super(0,sb,data,silent);
}

static DECLARE_FSTYPE(sffs_fs_type, "sffs", sffs_read_super, FS_REQUIRES_DEV);
#endif

#ifdef CONFIG_SFFS_RAM_ENABLED

static struct super_block *sffs_ram_read_super(struct super_block * sb, void * data, int silent)
{
	
	return sffs_internal_read_super(1,sb,data,silent);
}

static DECLARE_FSTYPE(sffs_ram_fs_type, "sffsram", sffs_ram_read_super, FS_SINGLE);
#endif // CONFIG_SFFS_RAM_ENABLED


static struct proc_dir_entry *my_proc_entry;
static struct proc_dir_entry *my_proc_ram_write_entry;

static char * sffs_dump_dev(char *buf,sffs_Device *dev,char *name)
{
	int blockNum;
	sffs_BlockInfo * ScanA;
/*	
	buf +=sprintf(buf,"\nDevice %s\n",name);
	buf +=sprintf(buf,"startBlock......... %d\n",dev->startBlock);
	buf +=sprintf(buf,"endBlock........... %d\n",dev->endBlock);
	buf +=sprintf(buf,"chunkGroupBits..... %d\n",dev->chunkGroupBits);
	buf +=sprintf(buf,"chunkGroupSize..... %d\n",dev->chunkGroupSize);
	buf +=sprintf(buf,"nErasedBlocks...... %d\n",dev->nErasedBlocks);
	buf +=sprintf(buf,"nTnodesCreated..... %d\n",dev->nTnodesCreated);
	buf +=sprintf(buf,"nFreeTnodes........ %d\n",dev->nFreeTnodes);
	buf +=sprintf(buf,"nObjectsCreated.... %d\n",dev->nObjectsCreated);
	buf +=sprintf(buf,"nFreeObjects....... %d\n",dev->nFreeObjects);
	buf +=sprintf(buf,"nFreeChunks........ %d\n",dev->nFreeChunks);
	buf +=sprintf(buf,"nPageWrites........ %d\n",dev->nPageWrites);
	buf +=sprintf(buf,"nPageReads......... %d\n",dev->nPageReads);
	buf +=sprintf(buf,"nBlockErasures..... %d\n",dev->nBlockErasures);
	buf +=sprintf(buf,"nGCCopies.......... %d\n",dev->nGCCopies);
	buf +=sprintf(buf,"garbageCollections. %d\n",dev->garbageCollections);
	buf +=sprintf(buf,"passiveGCs......... %d\n",dev->passiveGarbageCollections);
	buf +=sprintf(buf,"nRetriedWrites..... %d\n",dev->nRetriedWrites);
	buf +=sprintf(buf,"nRetireBlocks...... %d\n",dev->nRetiredBlocks);
	buf +=sprintf(buf,"eccFixed........... %d\n",dev->eccFixed);
	buf +=sprintf(buf,"eccUnfixed......... %d\n",dev->eccUnfixed);
	buf +=sprintf(buf,"tagsEccFixed....... %d\n",dev->tagsEccFixed);
	buf +=sprintf(buf,"tagsEccUnfixed..... %d\n",dev->tagsEccUnfixed);
	buf +=sprintf(buf,"cacheHits.......... %d\n",dev->cacheHits);
	buf +=sprintf(buf,"nDeletedFiles...... %d\n",dev->nDeletedFiles);
	buf +=sprintf(buf,"nUnlinkedFiles..... %d\n",dev->nUnlinkedFiles);
	buf +=sprintf(buf,"nBackgroudDeletions %d\n",dev->nBackgroundDeletions);
	buf +=sprintf(buf,"useNANDECC......... %d\n",dev->useNANDECC);
*/	

    buf += sprintf(buf,"::::: wearleveling :");
    for(blockNum = dev->startBlock; blockNum <= dev->endBlock; blockNum++)
	{
		ScanA = sffs_GetBlockInfo(dev, blockNum);
		
			buf +=sprintf(buf,"%d", ScanA->nErased);
	}
	buf +=sprintf(buf,"*");	

	
	return buf;	
}

static int  sffs_proc_read(
        char *page,
	char **start,
	off_t offset,
	int count,
	int *eof,
	void *data
	)
{

	char my_buffer[5000];
	char *buf;
    memset(my_buffer,-1,sizeof(my_buffer));
	buf = my_buffer;

	if (offset > 0) return 0;

	/* Fill the buffer and get its length */
	buf +=sprintf(buf,"SFFS built:"__DATE__ " "__TIME__"\n%s\n%s\n", sffs_fs_c_version,sffs_guts_c_version);
	
	if(sffs_dev) buf = sffs_dump_dev(buf,sffs_dev,"sffs");
	if(sffsram_dev) buf = sffs_dump_dev(buf,sffsram_dev,"sffsram");
	

	strcpy(page,my_buffer);
	return strlen(my_buffer);
}


static int  sffs_proc_ram_write(
        char *page,
	char **start,
	off_t offset,
	int count,
	int *eof,
	void *data
	)
{

	return count;
}


static int init()
{
    int i;
    g_nStartRead=0;
    g_nEndRead=0;
    memset(g_strCreate,0,sizeof(g_strCreate));
    memset(g_strCreateMount,0,sizeof(g_strCreateMount));    
    g_nFileCreate=0;
    g_nIsCrash=0;
    g_nCrashEnd=0;
    g_nCrashStart=0;
    g_nTransBlock=0;
    g_nTranCnt=0;
    g_nCacheCnt=0;
    g_nTranFlag=0;
    g_nCreateflag=0;
    g_tNode=NULL;
    g_cache.chunkId=0;
    memset(g_cache.data,0,512);
    g_cache.dirty=0;
    g_cache.lastUse=0;
    g_cache.nBytes=0;
    g_cache.object=NULL;
    g_cache.previd=0;
    g_cache.idInChunk=0;

    
}

// ó εɴ ȣ.module_init(init_sffs_fs)
static int __init init_sffs_fs(void)
{
	int error = 0;

    init();        
	sffs_dev = sffsram_dev = NULL;
	

	//procfs ؼ  ִ .

    my_proc_entry = create_proc_read_entry("sffs",
                                           S_IRUGO | S_IFREG,
					   &proc_root,
					   sffs_proc_read,
					   NULL);
    if(!my_proc_entry)
    {
       return -ENOMEM;
    }

#ifdef CONFIG_SFFS_RAM_ENABLED
#if 0
    my_proc_ram_write_entry = create_proc_entry("sffs_ram",
                                           S_IRUGO | S_IFREG,
					   &proc_root);
 
    if(!my_proc_ram_write_entry)
    {
       return -ENOMEM;
    }
    else
    {
    	my_proc_ram_write_entry->write_proc = sffs_proc_ram_write;
    }
#endif

    //vfs  ý .
    error = register_filesystem(&sffs_ram_fs_type);
    if(error)
    {
    	return error;
    }
#endif //CONFIG_SFFS_RAM_ENABLED

#ifdef CONFIG_SFFS_MTD_ENABLED
	error = register_filesystem(&sffs_fs_type);
	if(error)
	{
#ifdef CONFIG_SFFS_RAM_ENABLED
		unregister_filesystem(&sffs_ram_fs_type);
#endif //CONFIG_SFFS_RAM_ENABLED
	}
#endif // CONFIG_SFFS_MTD_ENABLED

	return error;
}

static void __exit exit_sffs_fs(void)
{
    remove_proc_entry("sffs",&proc_root);
    
#ifdef CONFIG_SFFS_RAM_ENABLED
	unregister_filesystem(&sffs_ram_fs_type);
#endif
#ifdef CONFIG_SFFS_MTD_ENABLED
	unregister_filesystem(&sffs_fs_type);
#endif

}

module_init(init_sffs_fs)
module_exit(exit_sffs_fs)

MODULE_DESCRIPTION("SFFS - a NAND specific flash file system");
MODULE_AUTHOR("Charles Manning, Aleph One Ltd., 2002");
MODULE_LICENSE("GPL");

